// Routines for accessing the AudioTron
// Copyright (c) 2003 by Tom Grandgent

#include "acwAT.h"
#include "acwGlobals.h"
#include "base64.h"

// Given a hostname, returns the IP address - returns 0 on error
unsigned int resolveHostname(char *hostname)
{
	HOSTENT *ht;				// HOSTENT struct used for DNS lookup results
	struct sockaddr_in addr;	// Address struct to be filled in from HOSTENT struct

	// Resolve the given hostname into an IP address
	ht = gethostbyname(hostname);
	if (!ht)
	{
		Log("[resolveHostname] ERR: Unable to resolve '%s', Winsock error %d", hostname, WSAGetLastError());
		return 0;
	}

	// Take the first address in the list (note that there can be multiple ones, but the first is all you need)
	memcpy(&(addr.sin_addr), ht->h_addr_list[0], sizeof(addr.sin_addr));
	if (addr.sin_addr.s_addr == INADDR_ANY)
	{
		Log("[resolveHostname] ERR: Unable to resolve '%s'", hostname);
		return 0;
	}

	return addr.sin_addr.s_addr;
}

// Performs an HTTP request - returns -1 on error, otherwise the # of bytes in the output
int httpRequest(char *server, int port, char *path, char *username, char *password, char *output, int outputLen)
{
	SOCKET sock;				// The socket!
	struct sockaddr_in addr;	// Address struct to be filled in for the connect() call
	int i;
	int bytecount = 0;
	int failure = 0;
	char buf[4096];

	if (!output)
		return -1;

	*output = '\0';

	// Resolve the hostname to an IP address
	addr.sin_addr.s_addr = resolveHostname(server);
	if (!addr.sin_addr.s_addr)
		return 1;

	// Allocate a socket
	sock = INVALID_SOCKET;
	if ((sock = socket(PF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET)
	{
		Log("[httpRequest] ERR: Unable to allocate socket");
		return 1;
	}

	// Connect the socket to the web server
	addr.sin_port = htons(port);
	addr.sin_family = PF_INET;
	if (connect(sock, (LPSOCKADDR) &addr, sizeof(struct sockaddr)) == SOCKET_ERROR)
	{
		i = WSAGetLastError();
		Log("[httpRequest] ERR: Unable to connect, error %d", i);
		closesocket(sock);
		sock = INVALID_SOCKET;
		return 1;
	}

	// Form the HTTP request
	sprintf(buf, 
		"GET %s HTTP/1.0\r\n"
		"Accept: */*\r\n", path);

	// Supply credentials if necessary
	if (username[0] || password[0])
	{
		char authtext[256], authbase64[256];
		sprintf(authtext, "%s:%s", username, password);
		i = to64(authtext, authbase64, strlen(authtext));
		authbase64[i - 1] = '\0';
		sprintf(buf + strlen(buf), 
		"Authorization: Basic %s\r\n", authbase64);
	}

	// Add the blank line signalling the end of the headers
	strcat(buf, "\r\n");

	// Send the HTTP request
	i = send(sock, buf, strlen(buf), 0);
	if (i != SOCKET_ERROR)
	{
		// Get the response
		bytecount = 0;
		while (1)
		{
			// Receive a chunk of data, no larger than the buffer
			i = recv(sock, buf, sizeof(buf) - 1, 0);
			if (i != SOCKET_ERROR)
			{
				// If recv() returns 0, the connection has been closed gracefully.  Break out.
				if (!i)
					break;
				// Add the # of bytes received to the total # of bytes received
				bytecount += i;
				// Terminate the buffer with a null character so it can be safely used as a string
				buf[i] = '\0';
				// Add this chunk of received data to the output buffer, checking for overflow
				if (bytecount >= outputLen)
				{
					// Figure out how many bytes we can safely copy
					int maxBytesToCopy = outputLen - (bytecount - i) - 1;
					if (maxBytesToCopy > 0)
					{
						// Copy the bytes, and null-terminate the string
						int oldlen = strlen(output);
						strncat(output, buf, maxBytesToCopy);
						output[oldlen + maxBytesToCopy] = '\0';
					}
					Log("[httpRequest] Warning - more data was received than could be stored in the output buffer");
				}
				else
					strcat(output, buf);
			}
			else
			{
				Log("[httpRequest] recv() failed, Winsock error is %d.", WSAGetLastError());
				failure = 1;
				break;
			}
		}
	}
	else
	{
		// send() failed when trying to send HTTP request
		Log("[httpRequest] send() failed, Winsock error is %d.", WSAGetLastError());
		failure = 1;
	}

	// Close the socket
	closesocket(sock);

	if (failure)
		return -1;

	return bytecount;
}

// Return the current status of the specified AudioTron (placed in status) - returns 0 on success, 1 on error
int ATGetStatus(char *server, char *status, int statusLen)
{
	//Log("[ATGetStatus] Retrieving status from %s", server);

	if (httpRequest(server, 80, "/apigetstatus.asp", "", prefATPassword, status, statusLen) == -1)
	{
		Log("[ATGetStatus] Unable to get status from %s", server);
		return 1;
	}

	return 0;
}

// Goes through a string and replaces all instances of "origChar" with "newChar"
void strrepch(char *str, char origChar, char newChar)
{
	char *ch = str;
	while (*ch)
	{
		if (*ch == origChar)
			*ch = newChar;
		ch++;
	}
}

// Converts a value from 0 to 15 to a hex digit character (0-9, A-F)
char hexDigit(unsigned char ch)
{
	if (ch < 10)
		return '0' + ch;
	else
		return 'A' + ch - 10;
}

// Returns a pointer to a dynamically allocated string which is the URL-encoded input string
char *URLEncode(char *str)
{
	// Allocate a string that is long enough such that every single input character 
	// could be encoded with three bytes
	char *enc = (char *) malloc(strlen(str) * 3 + 1);

	char *chin = str;
	char *chout = enc;
	while (*chin)
	{
		// Replace spaces with +'s
		if (*chin == ' ')
			*chout++ = '+';
		else
		{
			// Allow letters and numbers
			if (*chin >= 'A' && *chin <= 'Z' ||
				*chin >= 'a' && *chin <= 'z' ||
				*chin >= '0' && *chin <= '9')
			{
				*chout++ = *chin;
			}
			else
			{
				// Encode other characters
				*chout++ = '%';
				*chout++ = hexDigit(*chin >> 4);
				*chout++ = hexDigit(*chin & 0x0F);
			}
		}
		// Move on to the next character
		chin++;
	}
	// Null-terminate the encoded string and reteurn it
	*chout = '\0';

	return enc;
}

// Sets a message on the display of the AudioTron - returns 0 on success, 1 on error
// Specify -1 for timeoutSeconds to avoid including the timeout parameter in the request
int ATSetMessage(char *server, char *msgLine1, char *msgLine2, int timeoutSeconds)
{
	char *path;
	char *line1, *line2;
	char output[512];

	// Make a URL-encoded copy of the first line
	if (msgLine1 && msgLine1[0])
		line1 = URLEncode(msgLine1);
	else
		line1 = strdup("+");	// blank line

	// Make a URL-encoded copy of the second line
	if (msgLine2 && msgLine2[0])
		line2 = URLEncode(msgLine2);
	else
		line2 = strdup("+");	// blank line

	// Form the request path - http://<ipaddr>/apimsg.asp?line1=<text>&line2=<text>&timeout=<sec>
	path = (char *) malloc(strlen(line1) + strlen(line2) + 64);
	sprintf(path, "/apimsg.asp?line1=%s&line2=%s", line1, line2);
	if (timeoutSeconds != -1)
		sprintf(path + strlen(path), "&timeout=%d", timeoutSeconds);

	// Make the request
	if (httpRequest(server, 80, path, "", prefATPassword, output, sizeof(output)) == -1)
	{
		Log("[ATSetMessage] Unable to set message on %s", server);
		return 1;
	}

	// Free allocated strings
	free(path);
	free(line1);
	free(line2);

	return 0;
}

// Executes the given command on the AudioTron using apicmd.asp - returns 0 on success, 1 on error
int ATCommand(char *server, char *cmd)
{
	char path[256], output[256];

	if (strlen(cmd) > 200)
		return 1;
	sprintf(path, "/apicmd.asp?cmd=%s", cmd);

	// Make the request
	if (httpRequest(server, 80, path, "", prefATPassword, output, sizeof(output)) == -1)
	{
		Log("[ATCommand] Unable to execute command on %s", server);
		return 1;
	}

	return 0;
}

// Adds the given file to the AudioTron's play queue
int ATQueueFile(char *server, char *fileType, char *fileName)
{
	char path[512], output[256];

	if (strlen(fileName) + strlen(fileType) > 300)
		return 1;

	char *urlEncodedFileName = URLEncode(fileName);
	sprintf(path, "/apiqfile.asp?type=%s&file=%s", fileType, urlEncodedFileName);
	free(urlEncodedFileName);

	// Make the request
	if (httpRequest(server, 80, path, "", prefATPassword, output, sizeof(output)) == -1)
	{
		Log("[ATQueueFile] Unable to queue file on %s", server);
		return 1;
	}

	return 0;
}