// GUI code and program entrypoint for ACWatch
// Copyright (c) 2003 by Tom Grandgent

#include "acwatch.h"
#include "acwGlobals.h"		// Global variables

// Main window: WM_CREATE
BOOL Main_OnCreate(HWND hwnd, LPCREATESTRUCT lpCreateStruct)
{
	// Initialize the system tray icon
	NOTIFYICONDATA nid;
	nid.cbSize = sizeof(NOTIFYICONDATA);
	nid.hWnd = hwnd;
	nid.uID = 1; 
	nid.uFlags = NIF_MESSAGE | NIF_ICON | NIF_TIP;      
	nid.uCallbackMessage = UWM_SYSTRAY;
	nid.hIcon = (struct HICON__ *) LoadImage(hInstance, MAKEINTRESOURCE(IDI_MAIN), IMAGE_ICON,
						 GetSystemMetrics(SM_CXSMICON),
						 GetSystemMetrics(SM_CYSMICON), 0);
	strncpy(nid.szTip, "acwatch", 64);
	if (!Shell_NotifyIcon(NIM_ADD, &nid))
		Log("[WinMain] Shell_NotifyIcon() failed to initialize the tray icon!");

	// Create a listbox to hold the log entries
	RECT r;
	GetClientRect(hwnd, &r);
	hwndLBLog = CreateWindow("LISTBOX", "listbox", WS_CHILD | WS_VISIBLE | WS_VSCROLL | LBS_NOINTEGRALHEIGHT, 
		0, 0, r.right, r.bottom, hwnd, NULL, hInstance, NULL);
	HDC hdc = GetDC(hwnd);
	HFONT lbFont = CreateFont(-MulDiv(8, GetDeviceCaps(hdc, LOGPIXELSY), 72), 
		0, 0, 0, FW_DONTCARE, FALSE, FALSE, FALSE, 
		DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH, 
		"Arial");
	SendMessage(hwndLBLog, WM_SETFONT, (WPARAM) lbFont, TRUE);
	ReleaseDC(hwnd, hdc);

	// Post the WM_POST_CREATION message to do stuff after the window has been created
	PostMessage(hwnd, WM_POST_CREATION, 0, 0);

	return TRUE;
}

// Main window: WM_CLOSE
void Main_OnClose(HWND hwnd)
{
	static int closeTime = 0;

	Log("[Main_OnClose] Trying to exit");

	// Try to terminate the monitoring thread gracefully
	if (WaitForSingleObject((HANDLE) threadMonitor, 0) != WAIT_OBJECT_0)
	{
		// Signal the termination of the monitoring thread
		Log("[Main_OnClose] Signalling monitoring thread to stop");
		SetEvent(quitMonitor);
		Sleep(50);

		// If we've been waiting a while, just give up on the thread terminating gracefully
		if (!closeTime)
			closeTime = timeGetTime();
		else
		{
			if (timeGetTime() - closeTime > 2000)
			{
				Log("Warning: Monitoring thread didn't terminate in time.. exiting anyway");
				DestroyWindow(hwnd);
				return;
			}
		}

		PostMessage(hwnd, WM_CLOSE, 0, 0);	// Try again if the thread was still active...
		return;
	}

	Log("[Main_OnClose] Exiting");

	DestroyWindow(hwnd);
}

// Main window: WM_DESTROY
void Main_OnDestroy(HWND hwnd)
{
	// Remove the icon from the tray
	NOTIFYICONDATA nid;
	nid.cbSize = sizeof(NOTIFYICONDATA);
	nid.hWnd = hwnd;
	nid.uID = 1;
	nid.uFlags = 0;
	Shell_NotifyIcon(NIM_DELETE, &nid);

	// Exit the program
	PostQuitMessage(0);
}

// Main window: WM_POST_CREATION (custom message)
void Main_OnPostCreation(HWND hwnd)
{
	// Start the monitoring thread
	unsigned int tid;
	threadMonitor = _beginthreadex(NULL, 0, monitor_thread, NULL, 0, &tid);
	if (!threadMonitor)
		Log("ERROR: Unable to create monitoring thread!");
}

// Display the main window
void Main_openWindow(void)
{
	// Display the main window
	bMinimized = FALSE;
	ShowWindow(hwndApp, SW_RESTORE);
	SetForegroundWindow(hwndApp);

	// Clear the list box
	SendMessage(hwndLBLog, LB_RESETCONTENT, 0, 0);

	// Fill the list box with lines from the log file
	char buf[512];
	FILE *fp = fopen(logFilename, "r");
	if (fp)
	{
		while (fgets(buf, 512, fp))
		{
			if (buf[0])
				buf[strlen(buf) - 1] = '\0';
			SendMessage(hwndLBLog, LB_ADDSTRING, 0, (LPARAM) buf);
		}
		fclose(fp);

		// Select the last item in the list box to scroll to the bottom
		int itemCount = SendMessage(hwndLBLog, LB_GETCOUNT, 0, 0);
		SendMessage(hwndLBLog, LB_SETCURSEL, (WPARAM) itemCount - 1, 0);
		// Remove the selection
		SendMessage(hwndLBLog, LB_SETCURSEL, (WPARAM) -1, 0);
	}
}

// Main window: UWM_SYSTRAY (custom message) - a systray event has occurred
int Main_OnSysTray(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
	int ret;

	switch (lParam)
	{
		// The right button has been released over our icon
		case WM_RBUTTONUP:
			// Load a context menu
			HMENU hmenu, hpopup;
			MENUITEMINFO mii;
			POINT pt;
			GetCursorPos(&pt);
			hmenu = LoadMenu(hInstance, MAKEINTRESOURCE(IDM_TRAY));

			// Set the default menu item (which makes it bold)
			mii.cbSize = sizeof(MENUITEMINFO);
			mii.fMask = MIIM_STATE;
			mii.fState = MFS_DEFAULT;
			SetMenuItemInfo(hmenu, IDM_CONTEXT_ACWATCH, FALSE, &mii);

			// Display the context menu
			hpopup = GetSubMenu(hmenu, 0);
			SetForegroundWindow(hwnd);
			ret = TrackPopupMenu(hpopup, TPM_RETURNCMD | TPM_RIGHTBUTTON, pt.x, pt.y, 0, hwnd, NULL);
			DestroyMenu(hmenu);
			switch (ret)
			{
				case IDM_CONTEXT_EXIT: 
					PostMessage(hwnd, WM_CLOSE, 0, 0); 
					break;

				case IDM_CONTEXT_ACWATCH: 
					Main_openWindow();
					break;
			}
			break;

		// The left button has been released over our icon
		case WM_LBUTTONUP:
			// Show the main window
			Main_openWindow();
			break;
	}

	return TRUE;
}

// Main window: WM_SIZE
void Main_OnSize(HWND hwnd, UINT state, int cx, int cy)
{
	if (state == SIZE_MINIMIZED)
	{
		bMinimized = TRUE;

		// Clear the list box
		SendMessage(hwnd, LB_RESETCONTENT, 0, 0);

		// A minimize has been requested.  Hide the window completely.
		ShowWindow(hwnd, SW_HIDE);
	}

	MoveWindow(hwndLBLog, 0, 0, cx, cy, TRUE);
}

// Main window: WM_COMMAND
void Main_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
	switch (id)
	{
		case IDM_FILE_EXIT:
			PostMessage(hwnd, WM_CLOSE, 0, 0);
			break;
	}
}

// Main window procedure, processes window messages
LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
		HANDLE_MSG(hwnd, WM_CREATE, Main_OnCreate);
		HANDLE_MSG(hwnd, WM_CLOSE, Main_OnClose);
		HANDLE_MSG(hwnd, WM_DESTROY, Main_OnDestroy);
		HANDLE_MSG(hwnd, WM_SIZE, Main_OnSize);
		HANDLE_MSG(hwnd, WM_COMMAND, Main_OnCommand);
		case WM_POST_CREATION: Main_OnPostCreation(hwnd); return 0;
		case UWM_SYSTRAY: return Main_OnSysTray(hwnd, wParam, lParam);
	}

	return DefWindowProc(hwnd, msg, wParam, lParam);
}

// Registers window classes for windows created by this program
int RegisterWindowClasses(void)
{
	WNDCLASSEX wcl;

	// Initialize the main window class
	wcl.cbSize = sizeof(WNDCLASSEX);
	wcl.style = CS_HREDRAW | CS_VREDRAW;
	wcl.lpfnWndProc = WndProc;
	wcl.cbClsExtra = 0;
	wcl.cbWndExtra = 0;
	wcl.hInstance = hInstance;
	wcl.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_MAIN));
	wcl.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcl.hbrBackground = 0;
	wcl.lpszMenuName = MAKEINTRESOURCE(IDM_MAIN);
	wcl.lpszClassName = mainClassName;
	wcl.hIconSm = (HICON) LoadImage(hInstance, MAKEINTRESOURCE(IDI_MAIN), IMAGE_ICON, 
					GetSystemMetrics(SM_CXSMICON),
					GetSystemMetrics(SM_CYSMICON), 0);
	if (!RegisterClassEx(&wcl))
		return 1;

	return 0;
}

// Program entrypoint
int WINAPI WinMain(HINSTANCE hCurInstance, HINSTANCE hPrevInstance, LPSTR lpszCmdLine, int nCmdShow)
{
	// Save the hInstance
	hInstance = hCurInstance;

	// Save the command line
	cmdLine = strdup(lpszCmdLine);

	// Prevent multiple instances
	if (FindWindow(mainClassName, NULL))
	{
		MessageBox(NULL, "Another instance of acwatch is already running!", "No multiple instances allowed", MB_ICONERROR);
		return 1;
	}

	// Get the path of where the program is running
	GetModuleFileName(GetModuleHandle(NULL), appPath, sizeof(appPath));
	char *slash = strrchr(appPath, '\\');
	if (slash)
		slash[1] = '\0';
	// Set up the log filename
	strcpy(logFilename, appPath);
	strcat(logFilename, "acwatch.log");
	// Delete the old log file
	unlink(logFilename);
	// Log program startup
	time_t now = time(NULL);
	Log("ACWatch compiled %s %s starting up at %s", __DATE__, __TIME__, asctime(localtime(&now)));

	// Read the config file
	readConfigFile();

	// Initialize Winsock
	init_winsock();

	// Register window classes
	if (RegisterWindowClasses())
	{
		MessageBox(NULL, "Error registering window classes!", "Initialization error", MB_ICONERROR);
		return 1;
	}

	// Create the main window
	int windowStyleEx = 0;
	hwndApp = CreateWindowEx(windowStyleEx, mainClassName, mainWindowTitle, 
			   WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, CW_USEDEFAULT, 
			   CW_USEDEFAULT, CW_USEDEFAULT, HWND_DESKTOP, NULL, hInstance, NULL);
	if (!hwndApp)
	{
		MessageBox(NULL, "Error creating main application window!", "Initialization error", MB_ICONERROR);
		return FALSE;
	}

	// Main program loop
	MSG msg;
	while (GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	// Shut down Winsock
	shutdown_winsock();

	return msg.wParam;
}